/* qsort.c                              */
/* POSIX version                        */

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/mman.h>
#include <stdio.h>
#include <stdlib.h>

int compare( const void *arg1, const void *arg2 );

int main(int argc, char *argv[])
{
        int hFile,hMap;
	int elems,size;
	void *pData;
        struct stat fdata;

	/* check arguments */
	if(argc<2)
	{
		printf("usage: %s file_to_sort\n",argv[0]);
		exit(-1);
	}
	/* open file for reading and writting */
	hFile=open(argv[1],O_RDWR);
	if(-1==hFile)
	{
            perror("Error at CreateFile");
            exit(1);
	}
        /* determine the size of file */
	if(-1==fstat(hFile,&fdata))
	{
            perror("Error at fstat");
            exit(-1);
	}
        
	size=fdata.st_size;
        
	/* check for valid number of integers */
	if(size%sizeof(int))
	{
		fprintf(stderr,"File contains an invalid number of Integers (size%%4 != 0)\n");
		exit(-1);
	}
	/* calculate the amount of numbers to sort */
	elems=size/sizeof(int);
       
	/* create the file mapping object */
	pData=mmap(0,size,PROT_READ|PROT_WRITE,MAP_SHARED,hFile,0);
	if(MAP_FAILED==pData)
	{
		
		perror("Error at mmap");
		exit(-1);
	}
              
        /* sort the numbers */
	qsort(pData,elems,sizeof(int),compare);

	/* write the result into the file */
	if(-1==msync(pData,size,MS_SYNC))
	{
            perror("Error at msync");
            exit(-1);
	}
               
	/* remove the mapped region */
	if(-1==munmap(pData,size))
	{
		perror("Error at munmap");
		exit(-1);
	}
	/* cleanup handles */
	if(-1==close(hFile))
	{
            perror("Error at CloseHandle");
            exit(-1);
	}
      	return 0;
}


/* compare 2 integers */
int compare( const void *arg1,  const void *arg2 )
{
	if( *((int*)arg1) == *((int*)arg2) )
		return 0; /*equal*/
	else 
		if( *((int*)arg1) > *((int*)arg2) )
			return 1;/*arg1 greater than arg2*/
		else
			return -1;/*arg1 less than arg2*/
}

