/* qsort.c                              */
/* Example solution to assignment 5.2.2 */
/* Windows Operating System Internals   */

#include <windows.h>
#include <stdio.h>
#include <stdlib.h>

int compare( const void *arg1, const void *arg2 );

void printErrorAndExit(const char *msg, DWORD err)
{
	LPSTR lpMsgBuf;
	if(FormatMessage(FORMAT_MESSAGE_ALLOCATE_BUFFER|
			          FORMAT_MESSAGE_FROM_SYSTEM | 
					  FORMAT_MESSAGE_IGNORE_INSERTS,
						NULL,
						err,
					  MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT), // Default language
					  (LPTSTR) &lpMsgBuf,
					  0,
					  NULL ))
	{
		fprintf(stderr,"%s : %s\n",msg,lpMsgBuf);
		LocalFree(lpMsgBuf);
	}
	else
	{
		fprintf(stderr,"Error at FormatMesage: %d\n",err=GetLastError());
	}
	exit(err);
}

int tsize;

int main(int argc, char *argv[])
{
	HANDLE hMap,hFile;
	DWORD err,size;
	int elems;
	LPVOID pData;

	/* check arguments */
	if(argc<3)
	{
		printf("usage: %s <size of type in byte> file_to_sort\n",argv[0]);
		exit(-1);
	}
    
    tsize=atoi(argv[1]);
    
	/* open file for reading and writting */
	hFile=CreateFile(argv[2],GENERIC_WRITE|GENERIC_READ,0,NULL,OPEN_EXISTING,FILE_ATTRIBUTE_NORMAL,NULL);
	if(INVALID_HANDLE_VALUE==hFile)
	{
		err=GetLastError();
		printErrorAndExit("Error at CreateFile",err);
	}
    /* determine the size of file */
	size=GetFileSize(hFile,NULL);
	if((err=GetLastError())!=NO_ERROR)
	{
		printErrorAndExit("Error at GetFileSize",err);
	}
	
	/* check for valid number of integers */
	if(size % tsize)
	{
		fprintf(stderr,"File contains an invalid number of types (size%%%d != 0)\n",tsize);
		exit(-1);
	}
	/* calculate the amount of numbers to sort */
	elems=size/tsize;

	/* create the file mapping object */
	hMap=CreateFileMapping(hFile,NULL,PAGE_READWRITE,0,0,NULL);
	if(NULL==hMap)
	{
		printErrorAndExit("Error at CreateFileMapping",GetLastError());
	}
	/* map the whole file into the process context */
	pData=MapViewOfFile(hMap,FILE_MAP_WRITE,0,0,0);
	if(NULL==pData)
	{
		
		printErrorAndExit("Error at MapViewOfFile",GetLastError());
		
	}
    /* sort the numbers */
	qsort(pData,elems,tsize,compare);

	/* write the result into the file */
	if(!FlushViewOfFile(pData,0))
	{
		err=GetLastError();
		printErrorAndExit("Error at UnmapViewOfFile",err);
	}

	/* remove the mapped file */
	if(!UnmapViewOfFile(pData))
	{
		err=GetLastError();
		printErrorAndExit("Error at UnmapViewOfFile",err);
		exit(err);
	}
	/* cleanup handles */
	if(!CloseHandle(hMap) || !CloseHandle(hFile) )
	{
		err=GetLastError();
		printErrorAndExit("Error at CloseHandle",err);
	}
	return 0;
}


/* compare 2 integers */
int compare( const void *arg1, const void *arg2 )
{
    switch(tsize)
    {
        case 1:
            if( *((char*)arg1) == *((char*)arg2)) 
                return 0;
            else
                if( *((char*)arg1) > *((char*)arg2))
                    return 1;
                else
                    return -1;
            break;
        case 2:
            if( *((short*)arg1) == *((short*)arg2))
                return 0;
            else
                if( *((short*)arg1) > *((short*)arg2))
                    return 1;
                else
                    return -1;
            break;    
        case 4:
            if( *((int*)arg1) == *((int*)arg2))
                return 0;
            else
                if( *((int*)arg1) > *((int*)arg2))
                    return 1;
                else
                    return -1;
            break;    
        case 8:
            if( *((long long*)arg1) == *((long long*)arg2))
                return 0;
            else
                if( *((long long*)arg1) > *((long long*)arg2))
                    return 1;
                else
                    return -1;
            break;    
    }
}